{{
SleepAndWakeGpsTest.spin

This short program is based on GPS_leap_second_test.  It's a testing ground for ideas on how to sleep the
GPS and still keep track of time.    It spits out lots of debug messgaes that can be captured using
ReatTerm (with time stamps) to see how we're doing.


25 March 2011
Version: 1.0
University Park, PA
PGB

}}

CON ' System Clock Settings
' These have been removed and replaced with a boot-up sequence which will reset the GPS clock
' System clock settings - 81.92 MHz
   _clkmode = xtal1 + pll16x
   _xinfreq = 5_120_000


CON ' Normal constants

  SD_Basepin        =  0  ' basepin for SD card driver
  M_RESET           =  4  ' Reset line for DataFlash and

  RX_FROM_MACM      = 16  ' in-coming channel from MACM UART
  TX_TO_MACM        = 17  ' out-going channel to MACM UART --   CURRENTLY NOT USED!

  TX_TO_NMEA        = 17  ' out-going channel to NMEA UART
  RX_FROM_NMEA      = 18  ' in-coming channel from NMEA UART

  GPS_XBEE_RESET    = 19  ' GPS reset pin active high; flipped in hardware

  PPS_PIN           = 26            ' PPS pin
  FAKE_PPS_PIN      = 29
  OUTPUT_PIN        = 28   ' Heartbeat PIN

  CR                = 13      ' ASCII <CR>
  LF                = 10      ' ASCII <LF>
  DASH              = 45      ' ASCII -
  DOT               = 46      ' ASCII .
  COLON             = 58      ' ASCII :
  SPACE             = 32      ' ASCII <SPACE>
  TAB               = 09

  DEBUG             = 0
  NMEA              = 1
  MACM              = 2
  DEBUG_BAUD        = 38400
  MACM_BAUD         = 38400
  NMEA_BAUD         = 38400

  acq_ON            = $80
  acq_OFF           = $00

  SYS_CLOCK         = 81_920_000

  ONE_MS            = SYS_CLOCK/1000      ' how many clk cnts in 1 ms?  
  GPS_TIMEOUT       = 300                 ' number of attempts to get time lock
  PPS_WINDOW        = SYS_CLOCK - ONE_MS  ' 1/100 of a second 
  NMEA_WINDOW       = 200*ONE_MS          ' 16_384_000/81_920_000 is 0.2sec; give the NMEA 0.2 seconds to arrive
  MACM_WINDOW       = 500*ONE_MS          ' pick value that will put us near the middle of the second
  PPS_ADJUSTMENT    = 573                 ' number of clock cycles between arrival of the PPS and when I can assign it = 7uS

OBJ
'******************************************* Object to Include *************************************
  uarts    : "pcFullDuplexSerial4FC"               ' 1 COG for 3 serial ports
  Num      : "Numbers"                             ' Include Numbers object for writing numbers to debuginal
  gps      : "NavsyncCW25"                         ' no COG required

VAR
'******************************************* Variable declarations *********************************
  byte mainCogId, serialCogId, writeMacmCogId, clockCogId
  byte gpsValid, gpsInfoInUse

  word numFiles
  word index

  long stack1[90], stack2[90], stack3[90]
  long waitTime[5]
  long time, date, latitude, longitude, jday, ppsCnts, fixType, numSV, numBytes, delta
  long year, month, day, hour, minute, second ' contain global time stamp
  long t, dt, h, m, s, YY, MM, DD             ' Timekeeping variables
  long semID, cog                             ' Semaphore ID and cog variables
  long cntsAtPPS, currentTime, oldTime

'******************************************* Public and Private Methods ****************************
PUB MAIN | idx,  Rx_byte                     '' Main method

  RESET_GPS

  Num.Init
  uarts.Init

  uarts.AddPort(DEBUG,31,30,-1,-1,0,0,DEBUG_BAUD) 'Add DEBUG port
  uarts.AddPort(NMEA,rx_from_nmea,tx_to_nmea,-1,-1,0,%000000,NMEA_BAUD) 'Add GPS NMEA port

  serialCogId    := uarts.Start - 1                                           'Start the ports
  DELAY_MS(5000)
  uarts.str(DEBUG,string(CR,"Welcome aboard."))

  
  GPS.INIT(NMEA)                                        ' setup some things in GPS 
  GPS.CONFIG_GPS_MESSAGE                                ' Config rest of gps settings

  DELAY_MS(5000)

  clockCogId      := cognew(MANAGE_TIME,@stack1)      'Start the CLOCK Cog
  uarts.str(DEBUG,string(13,"Handling clock duties in cog:      "))
  uarts.dec(DEBUG,clockCogId)

  repeat

  idx:=0
  repeat 5
'    uarts.str(NMEA,string("$PRTHQ,ITIM",CR,LF))
'    junk := cnt                                     
'    repeat while (cnt - junk < clkfreq * 3)             ' echo the NMEA port for 10 seconds
'      Rx_byte := uarts.rxcheck(NMEA)                    ' collect byte from MACM port
'      if Rx_byte <> -1                                  ' -1 means no data; don't do anything
'        uarts.tx(DEBUG,Rx_byte)

'    uarts.str(NMEA,string("$PRTHQ,ILLH",CR,LF))
'    junk := cnt                                     
'    repeat while (cnt - junk < clkfreq * 3)             ' echo the NMEA port for 10 seconds
'      Rx_byte := uarts.rxcheck(NMEA)                    ' collect byte from MACM port
'      if Rx_byte <> -1                                  ' -1 means no data; don't do anything
'        uarts.tx(DEBUG,Rx_byte)

    repeat
      waitTime[idx] := WAIT_FOR_CLOCK_LOCK(12*60)
    until waitTime[idx] <> -1
    idx++ 
    uarts.str(DEBUG,string(CR,"Sleeping clock."))
    GPS.SLEEP(1*60*1000)

  idx:=0
  repeat 5
    uarts.str(DEBUG,string(CR,"waitTime["))
    uarts.dec(DEBUG,idx)
    uarts.str(DEBUG,string("]: "))
    uarts.dec(DEBUG,waitTime[idx++])

  uarts.str(DEBUG,string(CR,"DONE."))
      
PUB MANAGE_TIME | endOfSecond, distance, messageReceived   ' method to be run in unique cog; should read and parse NMEA line and get
' MACM data into buffer
'Process gps receive characters both MACM and NMEA
'minimize processing in this cog, i.e. dont send to serial ports e.g. debug
'dont call waitcnt, etc. Dont call blocking receives from serial ports - use rxcheck

  dira[28] := 1
  waitpeq(0, constant(|<PPS_pin), 0)    ' wait for pin to go low
  waitpne(0, constant(|<PPS_pin), 0)    ' wait for pin to go high
  cntsAtPPS := cnt - PPS_ADJUSTMENT     ' the 573 is a measured delay between the arrival of the PPS and how long
                                        ' it takes to get to the next instruction (7uS) 
  uarts.str(DEBUG,string(CR,"Made it here."))
  gpsValid := TRUE
  messageReceived := FALSE               ' use this to keep track of the end of NMEA messages


  repeat
    if gpsValid
      distance := cnt - cntsAtPPS
      if distance > ONE_MS AND distance < NMEA_WINDOW AND NOT messageReceived
        outa[28] := 1
       ' if we are between 1ms and 200ms after the PPS watch for NMEA messages?
       ' we should jump in here 1ms after the PPS and normally it takes until
       ' ~ 110ms after the PPS to get all the NMEA data and for the GPS routines
       ' to parse it.  From if statement to the end completion of POPULATE_GPS_INFO takes
       ' ~100ms
        if (GPS.PROCESS_NMEA_MESSAGES == %10 )    ' have we collected 2 complete messages?
          messageReceived := TRUE                 ' indicate that we've gotten both messages
          outa[28] := 0
          POPULATE_GPS_INFO
          uarts.putc(DEBUG,CR)
          uarts.decx(DEBUG,time,6)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,hour,2)
          uarts.putc(DEBUG,COLON)
          uarts.decx(DEBUG,minute,2)
          uarts.putc(DEBUG,COLON)
          uarts.decx(DEBUG,second,2)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,date,6)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,year,4)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,month,2)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,day,2)
          uarts.putc(DEBUG,SPACE)
          uarts.str(DEBUG,GPS.LATITUDE_STRING)
          uarts.putc(DEBUG,SPACE)
          uarts.dec(DEBUG,latitude)
          uarts.putc(DEBUG,SPACE)
          uarts.str(DEBUG,GPS.LONGITUDE_STRING)
          uarts.putc(DEBUG,SPACE)
          uarts.dec(DEBUG,longitude)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,fixType,3)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,numSV,2)
          uarts.putc(DEBUG,SPACE)
          uarts.decx(DEBUG,cntsAtPPS,11)
          uarts.putc(DEBUG,SPACE)


      if ( distance => PPS_WINDOW )
        messageReceived := FALSE
       ' are we within 1/100 of a second of the next PPS?
        waitpeq(0, constant(|<PPS_pin), 0)    ' wait for pin to go low
        waitpne(0, constant(|<PPS_pin), 0)    ' wait for pin to go high
        cntsAtPPS := cnt - PPS_ADJUSTMENT
        endOfSecond  := cntsAtPPS + clkfreq   ' Current clock counter
    else
        ' do nothing



  return
   
PUB POPULATE_GPS_INFO ' method that goes out and gets all the pertinent gps info and puts it in expected var locations
'' this method takes about 250us and starts between 85ms-100ms after the PPS.  It does not start until the NMEA messages have
'' been parsed.  Including the latitude/longitude stuff increases this to 4.5ms.  

'  outa[28] := 1
  repeat while gpsInfoInUse    ' wait here while someone else uses this info
  gpsInfoInUse := TRUE

  time       := GPS.UTC_TIME_DECIMAL
  date       := GPS.UTC_DATE_DECIMAL
  jday       := 0'gps.jday(dateYear, dateMon, dateDay)

  year       := Date // 100 + 2000
  month      := Date/100 // 100
  day        := Date/10_000

  hour       := Time/10_000
  minute     := Time/100 // 100     
  second     := Time // 100

  latitude   := GPS.LATITUDE_DDMMMMMM
  longitude  := GPS.LONGITUDE_DDDMMMMMM       

  ppsCnts    := cntsAtPPS

  fixType    := GPS.FS_DECIMAL
  numSV      := GPS.NOSV_DECIMAL


  numBytes   := numBytes
  delta      := 1_000_000

  gpsInfoInUse := FALSE
'  outa[28] := 0

PUB WAIT_FOR_PPS | foundPPS, junk

    uarts.str(DEBUG,string(CR,"Looking for 1PPS:  "))
    foundPPS := FALSE
    repeat until foundPPS
      junk := cnt + 3*clkfreq
      repeat until(cnt > junk) OR foundPPS           ' allow 3 seconds to do this; then quit
        if ina[PPS_PIN] := 1
          foundPPS := TRUE
          uarts.str(DEBUG,string("Found 1PPS!"))
      if NOT foundPPS
        uarts.str(DEBUG,string("Rebooting GPS!"))
        RESET_GPS

PUB WAIT_FOR_CLOCK_LOCK(timeout) :attempt | clockLocked  

  WAIT_FOR_PPS

  uarts.str(DEBUG,string(CR,"Waiting for clock lock...TIMEOUT set to: "))
  uarts.dec(DEBUG,timeout)
  uarts.str(DEBUG,string(" seconds."))
  uarts.str(DEBUG,string(CR,"TIME   DATE   FIX  UTC_TOW  GPS_TOW ATTEMPT"))
  
  clockLocked := FALSE
  attempt     := 1

  repeat until clockLocked OR attempt == timeout
    if (GPS.PROCESS_NMEA_MESSAGES == %10 )
      if (GPS.UTC_TOW_DECIMAL <> GPS.GPS_TOW_DECIMAL) AND (GPS.FS_DECIMAL > 0)
        clockLocked := TRUE

      uarts.putc(DEBUG,DOT)
{      uarts.putc(DEBUG,CR)
      uarts.decx(DEBUG,GPS.UTC_TIME_DECIMAL,6)
      uarts.putc(DEBUG,SPACE)
      uarts.decx(DEBUG,GPS.UTC_DATE_DECIMAL,6)
      uarts.putc(DEBUG,SPACE)
      uarts.decx(DEBUG,GPS.FS_DECIMAL,3)
      uarts.putc(DEBUG,SPACE)
      uarts.putc(DEBUG,SPACE)
      uarts.decx(DEBUG,GPS.UTC_TOW_DECIMAL,6)
      uarts.putc(DEBUG,SPACE)
      uarts.putc(DEBUG,SPACE)
      uarts.putc(DEBUG,SPACE)
      uarts.decx(DEBUG,GPS.GPS_TOW_DECIMAL,6)
      uarts.putc(DEBUG,SPACE)
      uarts.putc(DEBUG,SPACE)
      uarts.decx(DEBUG,attempt,6)
}
      attempt++
      DELAY_MS(500)
    else
      uarts.putc(DEBUG,COLON)
      DELAY_MS(500)
      
  if clockLocked 
    uarts.str(DEBUG,string(13,"Time locked!"))
    gpsValid := TRUE
  else
    uarts.str(dEBUG,string(13,"TIMEOUT!"))
    RESET_GPS
    attempt := -1

  return attempt

{
PUB SetTimer(_year, _month, _day, _hour, _minute, _second)
{{Set the timer.
  Parameters:
    year       - year count
    month      - month count
    day        - day count
    hour       - hour count
    minute     - minute count
    seconds    - seconds count
}}

  repeat until not lockset(semID)         ' Set lock on time variables
  longmove(@year, @_year, 6)                  ' Copy values received
  lockclr(semID)                          ' Clear lock on time variables

PUB GetTime(yearAddr)                      '' Get current timestamp
{{Get the current timestamp.
  Parameters:
    hourAddr      - address of the caller's hour variable.

      IMPORTANT:   Caller must declare long variables in this sequence:
                   year, month, day, hour, minute, seconds
}}

  repeat until not lockset(semID)         ' Set lock on time variables 
  longmove(yearAddr, @year, 6)               ' Copy times to caller's vars
  lockclr(semID)                          ' Clear lock on time variables

PRI Timer

  t  := cnt                               ' Current clock counter
  dt := clkfreq                           ' Ticks in 1 second

  repeat                                  ' Infinite loop
    waitcnt(t+=dt)                        ' Wait for the next second
    repeat until not lockset(semID)       ' Set lock
    s++                                   ' Add 1 to second count
    if s == 60                            ' If seconds = 60
      s := 0                              ' Set seconds to 0
      m++                                 ' Increment minutes
    if m == 60                            ' if minutes = 60
      m := 0                              ' Set minutes to 0
      h++                                 ' Increment hours
    if h == 24                            ' If hours = 24
      h := 0                              ' Set hours to 0
      DD++                                ' Increment days

      if (DD == 31 AND (MM==04 OR MM==06 OR MM==09 OR MM==11))
      ' Have we reached the end of months with 30 days? Apr, Jun, Sep, or Nov?
        MM++                                ' Increment month
        DD := 1                             ' Set day of month to 1
      if (DD == 32 AND (MM==01 OR MM==03 OR MM==05 OR MM==07 OR MM==08 OR MM==10 OR MM==12))
      ' Have we reached the end of months with 31 days? Jan, Mar, Jul, Aug, Oct, and Dec
        MM++                                ' Increment month
        DD := 1                             ' Set day of month to 1
      if (DD == 30 AND MM==02 AND LEAP_YEAR)
      ' Have we reached the end of Feb during a leap year:
        MM++                                ' Increment month
        DD := 1                             ' Set day of month to 1
      if (DD == 29 AND MM==02 AND NOT LEAP_YEAR)
      ' Have we reached the end of Feb during a non-leap year?
        MM++                                ' Increment month
        DD := 1                             ' Set day of month to 1

    if (MM == 13)
      YY++                                ' Increment year
      MM := 1                             ' Set month to 1
      DD := 1                             ' Set day of month to 1

    lockclr(semID)                        ' Clear Lock

}
PUB DELAY_MS(mS)
  waitcnt(clkfreq/1000 * mS + cnt)
  return

PUB RESET_GPS
'' Setup_clock method sets the prop to use the internal RCFast clock.  It resets the GPS so
'' we know the state after powerup.  It then waits for ~5 seconds.  If it gets a 1PPS high
'' during those 5 seconds it will continue to boot normally.  If after 5 seconds it hasn't
'' seen a proper 1PPS, it will reboot itself and start this process again.


' RESET GPS
  dira[GPS_XBEE_RESET]~~      ' set GPS_XBEE_RESET to output
  outa[GPS_XBEE_RESET] := 1   ' set it high
  waitcnt(clkfreq + cnt)      ' wait ~1 second while the GPS boots
  outa[GPS_XBEE_RESET] := 0   ' set it low
  dira[GPS_XBEE_RESET]~       ' set to input

  return

PRI LEAP_YEAR
    if (YY // 4 == 0)
      if (YY // 100 == 0)
        if (YY // 400 == 0)
          return TRUE
        else
          return FALSE
      else
        return TRUE
    else
       return FALSE

PUB PASS_THROUGH (port) | Rx_byte
' infinite loop which simply reads from port and spits it to port 0 which should be the PST.

  repeat
    Rx_byte := uarts.rxcheck(port)          ' collect byte from MACM port
    if Rx_byte <> -1                     ' -1 means no data; don't do anything
      if port == NMEA
        uarts.putc(debug,Rx_byte)
      else
        uarts.hex(debug,Rx_byte,2)

  return

DAT
